from contextlib import asynccontextmanager
from typing import Annotated

import uvicorn
from fastapi import FastAPI, Path, status
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse

from portswatch.classes.logs import Log
from portswatch.classes.settings import Settings
from portswatch.controllers.ip import IPController
from portswatch.cruds.ip import IPCrud
from portswatch.cruds.port import PortCrud
from portswatch.cruds.ports_checker import PortCheckerCrud
from portswatch.cruds.unexpected_ports import UnexpectedPortCrud
from portswatch.models.ip import IPcreateModel, IPqueryModel
from portswatch.models.ports import PortsbaseModel, PortscreateModel, PortsqueryModel

settings = Settings()
log = Log(name=__name__)


@asynccontextmanager
async def lifespan(app: FastAPI):
    logger = Log.uvicorn_log()
    yield


app = FastAPI(lifespan=lifespan)


@app.get("/")
async def read_root():
    return {"Hello": "World"}


@app.post("/ip", status_code=status.HTTP_201_CREATED)
async def post_ip(ip: IPcreateModel | IPqueryModel):
    if not IPController(ip).is_ip_authorized():
        message = {
            "message": (
                f"{ip.ip} is not authorized by instance policy. "
                f"Instance authorize this type of IP: "
                f"{' | '.join(settings.authorized_ips_type)}"
            )
        }
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, content=message
        )
    if IPCrud().get_ip(ip):
        message = {"message": (f"{ip.ip} already exist. ")}
        return JSONResponse(status_code=status.HTTP_409_CONFLICT, content=message)

    Ip_crud = IPCrud().add(ip)
    if not Ip_crud:
        message = {"message": f"{ip.ip} isn't add into PortsWatch."}
        return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content=message)
    return JSONResponse(content={})


@app.get("/ip/{ip}", status_code=status.HTTP_200_OK)
async def read_ip(ip: Annotated[str, Path(title="The IP to get")]):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip = IPCrud().get_ip(ip_model)
    ports = []
    unexpected_ports = []
    check_ports = []
    for port in ip.ports:
        ports.append(port.port)
    for port in ip.unexpected_ports:
        unexpected_ports.append(port.port)
    for port in ip.check_ports:
        check_ports.append(port.port)
    ip.ports = ports
    ip.unexpected_ports = unexpected_ports
    ip.check_ports = check_ports
    message = jsonable_encoder(ip.dict())
    return JSONResponse(content=message)


@app.put("/ip/{ip}/check", status_code=status.HTTP_200_OK)
async def update_ip_check(ip: Annotated[str, Path(title="The IP to get")]):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip = IPCrud().update_check(ip_model, check="start")
    return JSONResponse(content={})


@app.delete("/ip/{ip}/delete", status_code=status.HTTP_204_NO_CONTENT)
async def delete_ip_check(ip_query: Annotated[str, Path(title="The IP to get")]):
    is_ip_global(ip_query)
    ip_model = IPqueryModel(ip=ip_query)
    ip = IPCrud().get_ip(ip=ip_model)
    IPCrud().finish_check(ip_model)
    UnexpectedPortCrud().delete_ports(ip)
    return JSONResponse(content={})


@app.delete("/ip/{ip}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_ip(ip: Annotated[str, Path(title="The IP to get")]):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    if not IPCrud().delete(ip_model):
        message = {"message": f"Cant't delete {ip}."}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    return JSONResponse(content={})


@app.post("/ip/{ip}/port", status_code=status.HTTP_201_CREATED)
async def write_port(
    ip: Annotated[str, Path(title="The IP attached")],
    ports: PortsbaseModel | PortsqueryModel,
):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip_data = IPCrud().get_ip(ip_model)
    if not ip_data:
        message = {"message": "Curious moment. IP is valid but ID is not"}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    port = PortscreateModel(port=ports.port, ip_id=ip_data.id)
    port_crud = PortCrud().get_port(port)
    if port_crud:
        message = {"message": f"Port {ports.port} already exist for {ip}."}
        return JSONResponse(status_code=status.HTTP_409_CONFLICT, content=message)
    port_crud = PortCrud().add(port)
    if not port_crud:
        message = {
            "message": f"Port {ports.port} (for ip {ip}) isn't add into PortsWatch."
        }
        return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content=message)
    return JSONResponse(content={})


@app.post("/ip/{ip}/check/port", status_code=status.HTTP_201_CREATED)
async def write_check_port(
    ip: Annotated[str, Path(title="The IP attached")],
    port: PortsbaseModel | PortsqueryModel,
):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip_data = IPCrud().get_ip(ip_model)
    if not ip_data:
        message = {"message": "Curious moment. IP is valid but ID is not"}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    port_model = PortscreateModel(port=port.port, ip_id=ip_data.id)
    port_crud = PortCheckerCrud().get_port(port_model)
    if port_crud:
        message = {"message": f"Port {port.port} already exist for {ip}."}
        return JSONResponse(status_code=status.HTTP_409_CONFLICT, content=message)
    port_crud = PortCheckerCrud().add(port_model)
    if not port_crud:
        message = {
            "message": f"Port {port.port} (for ip {ip}) isn't add into PortsWatch."
        }
        return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content=message)
    return JSONResponse(content={})


@app.get("/ip/{ip}/port/{port}", status_code=status.HTTP_200_OK)
async def read_port(
    ip: Annotated[str, Path(title="The IP to get")],
    port: Annotated[int, Path(title="The port")],
):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip_data = IPCrud().get_ip(ip_model)
    if not ip_data:
        message = {"message": "Curious moment. IP is valid but ID is not"}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    port = PortsqueryModel(port=port, ip_id=ip_data.id)
    port_crud = PortCrud().get_port(port)
    if not port_crud:
        message = {"message": f"Port {port.port} not for {ip}."}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=message)
    return JSONResponse(content=jsonable_encoder(port_crud.dict()))


@app.delete("/ip/{ip}/port/{port}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_port(
    ip: Annotated[str, Path(title="The IP to get")],
    port: Annotated[int, Path(title="The port")],
):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip_data = IPCrud().get_ip(ip_model)
    if not ip_data:
        message = {"message": "Curious moment. IP is valid but ID is not"}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    port = PortsqueryModel(port=port, ip_id=ip_data.id)
    port_crud = PortCrud().get_port(port)
    print(port_crud)
    if not port_crud:
        message = {"message": f"Port {port.port} not for {ip}."}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=message)
    if not PortCrud().delete(port_crud):
        message = {"message": f"Cant't delete {port.port} for {ip}."}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )

    return JSONResponse(content={})


@app.post("/ip/{ip}/unexpectedport", status_code=status.HTTP_201_CREATED)
async def post_unexpected_port(
    ip: Annotated[str, Path(title="The IP to get")],
    port: PortsbaseModel | PortsqueryModel,
):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip_data = IPCrud().get_ip(ip_model)
    if not ip_data:
        message = {"message": "Curious moment. IP is valid but ID is not"}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    port_model = PortscreateModel(port=port.port, ip_id=ip_data.id)
    port_crud = UnexpectedPortCrud().get_port(port_model)
    if port_crud:
        message = {"message": f"Port {port.port} already exist for {ip}."}
        return JSONResponse(status_code=status.HTTP_409_CONFLICT, content=message)
    port_crud = UnexpectedPortCrud().add(port_model)
    if not port_crud:
        message = {
            "message": f"Port {port.port} (for ip {ip}) isn't add into PortsWatch."
        }
        return JSONResponse(status_code=status.HTTP_400_BAD_REQUEST, content=message)
    return JSONResponse(content={})


@app.delete("/ip/{ip}/unexpectedport/{port}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_unexpected_port(
    ip: Annotated[str, Path(title="The IP to get")],
    port: Annotated[int, Path(title="The port")],
):
    is_ip_global(ip)
    ip_model = IPqueryModel(ip=ip)
    ip_data = IPCrud().get_ip(ip_model)
    if not ip_data:
        message = {"message": "Curious moment. IP is valid but ID is not"}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )
    port = PortsqueryModel(port=port, ip_id=ip_data.id)
    port_crud = UnexpectedPortCrud().get_port(port)
    print(port_crud)
    if not port_crud:
        message = {"message": f"Port {port.port} not for {ip}."}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=message)
    if not UnexpectedPortCrud().delete(port_crud):
        message = {"message": f"Cant't delete {port.port} for {ip}."}
        return JSONResponse(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, content=message
        )

    return JSONResponse(content={})


@app.get("/ips", status_code=status.HTTP_200_OK)
async def read_ips():
    message = {"ips": []}
    for ip in IPCrud().get_ips():
        print(ip)
        ports = []
        for port in ip.ports:
            ports.append(port.port)
        ip.ports = ports
        message["ips"].append(jsonable_encoder(ip.dict()))
    return JSONResponse(content=message)


@app.get("/ips/check", status_code=status.HTTP_200_OK)
async def get_ips_check():
    ips = IPCrud().get_ips_checking()
    return JSONResponse(content=jsonable_encoder(ips))


def is_ip_global(ip_query: str):
    check_ip_bool, check_ip_message = IPController.is_ip(ip_query)
    if not check_ip_bool:
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST, content=check_ip_message
        )

    ip_model = IPqueryModel(ip=ip_query)
    ip = IPCrud().get_ip(ip_model)
    if not ip:
        message = {"message": f"{ip_model.ip.compressed} not found"}
        return JSONResponse(status_code=status.HTTP_404_NOT_FOUND, content=message)


if __name__ == "__main__":
    uvicorn.run(
        "api:app",
        host=settings.host,
        port=settings.port,
        reload=settings.debug,
        log_config=None,
    )
