import ipaddress
import random
from multiprocessing import Pool
from os import geteuid
from time import sleep

import nmap3
import requests
import typer

from portswatch.classes.logs import Log
from portswatch.classes.settings import Settings

log = Log(name=__name__)
settings = Settings()
app = typer.Typer()
nmap = nmap3.NmapScanTechniques()


def get_ips_random(number_ip: int):
    try:
        r = requests.get(f"{settings.url}/ips")
        r.raise_for_status
    except requests.exceptions.HTTPError as error:
        error_code = error.response.status_code
        log.debug(error.response.headers)
        log.error(f"Can't get ips. Error code : {error_code}")
        return
    except requests.RequestException as e:
        log.error(str(e))
        return
    ips_registered = r.json()
    ips = []
    while len(ips) < number_ip:
        ip = random.choice(ips_registered["ips"])
        try:
            r = requests.put(f"{settings.url}/ip/{ip['ip']}/check", json={})
            r.raise_for_status
        except requests.exceptions.HTTPError as error:
            error_code = error.response.status_code
            log.debug(error.response.headers)
            log.error(f"Can't get ips. Error code : {error_code}")
            return
        except requests.RequestException as e:
            log.error(str(e))
            return
        ips.append(ip["ip"])
    return ips


def get_ips(number_ip: int):
    log.debug(f"Looking for {number_ip} ip(s) to check")
    ips = []
    try:
        r = requests.get(f"{settings.url}/ips/check")
        r.raise_for_status
    except requests.exceptions.HTTPError as error:
        error_code = error.response.status_code
        log.debug(error.response.headers)
        log.error(f"Can't get ips. Error code : {error_code}")
        return
    except requests.RequestException as e:
        log.error(str(e))
        return
    ips_checking = r.json()
    for ip in ips_checking:
        ips.append(ip["ip"])
    log.debug(f"Found {len(ips_checking)} ip(s) which match criteria in database")
    if len(ips_checking) < number_ip:
        log.debug(f"{len(ips_checking)} is lower than {number_ip}")
        number_ip = number_ip - len(ips_checking)
        log.debug("Reduce number_ip by the number of actual fetched IP")
        log.debug(f"New number_ip value : {number_ip}")
        log.debug("Fetching random IP registerd in API")
        ips_random = get_ips_random(number_ip=number_ip)
        ips.extend(ips_random)
        ips.extend(ips_checking)
    else:
        while len(ips) < number_ip:
            ip = random.choice(ips_checking)
            ips.extend(str(ip["ip"]))
    log.debug(f"Theses IP will be tested : {ips}")
    return ips


def get_ports(ips_query: list):
    ips = []
    for ip in ips_query:
        try:
            r = requests.get(f"{settings.url}/ip/{ip}")
            r.raise_for_status
        except requests.exceptions.HTTPError as error:
            error_code = error.response.status_code
            log.debug(error.response.headers)
            log.error(f"Can't get ips. Error code : {error_code}")
            return
        except requests.RequestException as e:
            log.error(str(e))
            return
        ip_api = r.json()
        ports_to_check = []
        while len(ports_to_check) < settings.number_check_ports:
            port = random.randrange(1, 65535)
            if port not in ip_api["check_ports"]:
                ports_to_check.append(port)
        ip_api["ports_to_check"] = ports_to_check
        ips.append(ip_api)
    return ips


def run_nmap(ip_ports):
    return nmap.nmap_tcp_scan(
        ip_ports["ip"],
        args=(
            f"-{ipaddress.ip_address(ip_ports['ip']).version} -Pn "
            f"{'-sU' if geteuid == 0 else ''} "
            f"-p {','.join(str(port) for port in ip_ports['ports_to_check'])}"
        ),
    )


def process_result(nmap_result: dict):
    ip = next(iter(nmap_result))
    log.debug(f"Get ports for {ip}")
    try:
        r = requests.get(f"{settings.url}/ip/{ip}")
        r.raise_for_status
    except requests.exceptions.HTTPError as error:
        error_code = error.response.status_code
        log.debug(error.response.headers)
        log.error(f"Can't get ips. Error code : {error_code}")
        return
    except requests.RequestException as e:
        log.error(str(e))
        return
    ips_checking = r.json()
    for port in nmap_result[ip]["ports"]:
        if port["portid"] in ips_checking["ports"]:
            log.info(f"{port['portid']} is allowed on {ip}, pass")
            continue
        if port["state"] not in settings.authorized_ports_state:
            log.warning(f"{port['portid']} is in state not allowed ({port['state']}) !")
            post_check_port(ip=ip, port=port["portid"], bad_state=True)
        post_check_port(ip=ip, port=port["portid"], bad_state=False)
    try:
        r = requests.get(f"{settings.url}/ip/{ip}")
        r.raise_for_status
    except requests.exceptions.HTTPError as error:
        error_code = error.response.status_code
        log.debug(error.response.headers)
        log.error(f"Can't get ips. Error code : {error_code}")
        return
    except requests.RequestException as e:
        log.error(str(e))
        return
    ip_checking = r.json()
    if len(ip_checking["check_ports"]) >= 65535:
        log.info(f"{ip} was succcessfuly scanned")
        try:
            r = requests.delete(f"{settings.url}/ip/{ip}/check")
            r.raise_for_status
        except requests.exceptions.HTTPError as error:
            error_code = error.response.status_code
            log.debug(error.response.headers)
            log.error(f"Can't get ips. Error code : {error_code}")
            return
        except requests.RequestException as e:
            log.error(str(e))
            return


def post_check_port(ip: str, port: int, bad_state: bool):
    if bad_state:
        uri = f"{settings.url}/ip/{ip}/unexpectedport"
    else:
        uri = f"{settings.url}/ip/{ip}/check/port"
    try:
        payload = {"port": port}
        r = requests.post(uri, json=payload)
        r.raise_for_status
    except requests.exceptions.HTTPError as error:
        error_code = error.response.status_code
        log.debug(error.response.headers)
        log.error(f"Can't get ips. Error code : {error_code}")
        return
    except requests.RequestException as e:
        log.error(str(e))
        return


def daemon_main():
    log.info(f"Getting {settings.number_parallel_ip} to test")
    ips = get_ips(number_ip=settings.number_parallel_ip)
    log.info("Get ports for selected IP")
    ips_ports = get_ports(ips_query=ips)
    with Pool(settings.number_parallel_ip) as p:
        log.info("Run nmap for selected IP and ports")
        nmap_results = []
        for ip_ports in ips_ports:
            nmap_results.append(p.map(run_nmap, [ip_ports])[0])
        log.info("Process results")
        for result in nmap_results:
            process_result(nmap_result=result)


@app.command()
def daemon():
    while True:
        daemon_main()
        log.info(f"Process finish. Sleep {settings.delay_between_check}")
        sleep(settings.delay_between_check)


if __name__ == "__main__":
    app()
