import sqlite3
from uuid import UUID, uuid4

import pendulum
from sqlalchemy import exc

from portswatch.classes.logs import Log
from portswatch.database import SessionLocal
from portswatch.models.ip import IPcreateModel, IPModel, IPqueryModel
from portswatch.schemas import IP

log = Log(name=__name__)


class IPCrud:
    def add(self, ip: IPcreateModel):
        ip_schema = IP(id=uuid4().bytes, ip=ip.ip.packed)
        with SessionLocal() as session, session.begin():
            try:
                session.add(ip_schema)
            except exc.SQLAlchemyError as e:
                log.debug(
                    f"Portswatch encounter an error when it try to add {ip.ip} : {e}"
                )
                session.rollback()
                return False
            except sqlite3.IntegrityError as e:
                log.debug(
                    f"Sqlite encounter a problem during the processing of {ip.ip} : {e}"
                )
                return False
            else:
                session.commit()
                return True

    def update_check(self, ip: IPqueryModel, check: str):
        with SessionLocal() as session, session.begin():
            ip_query = session.query(IP).filter(IP.ip == ip.ip.packed).first()
            if ip_query:
                ip_query.last_check_start = pendulum.now()
                ip_query.last_check_end = pendulum.yesterday()
                session.commit()
                return True
        return False

    def finish_check(self, ip: IPqueryModel, check: str):
        with SessionLocal() as session, session.begin():
            ip_query = session.query(IP).filter(IP.ip == ip.ip.packed).first()
            if ip_query:
                ip_query.last_check_end = pendulum.now()
                session.commit()
                return True
        return False

    def get_ip(self, ip: IPqueryModel):
        with SessionLocal() as session:
            ip_query = session.query(IP).filter(IP.ip == ip.ip.packed).first()
            if ip_query:
                return IPModel(
                    id=ip_query.id,
                    ip=ip_query.ip,
                    ports=ip_query.ports,
                    unexpected_ports=ip_query.unexpected_ports,
                    check_ports=ip_query.check_ports,
                )
            else:
                return None

    def get_ip_by_id(self, id: UUID):
        with SessionLocal() as session:
            ip = session.query(IP).filter(IP.id == id.bytes).first()
            if ip:
                return IPModel(id=ip.id, ip=ip.ip)
            else:
                return None

    def get_ips(self):
        with SessionLocal() as session:
            ips = []
            for ip in session.query(IP).all():
                ips.append(IPModel(id=ip.id, ip=ip.ip, ports=ip.ports))
            return ips

    def delete(self, ip: IPqueryModel):
        with SessionLocal() as session:
            ip_query = session.query(IP).filter(IP.ip == ip.ip.packed).first()
            if not ip_query:
                return None
        with SessionLocal() as session, session.begin():
            try:
                session.delete(ip_query)
            except exc.SQLAlchemyError as e:
                log.debug(
                    f"Portswatch encounter an error when it try to delete {ip.ip}: {e}"
                )
                session.rollback()
                return False
            except sqlite3.IntegrityError as e:
                log.debug(
                    f"Sqlite encounter a problem during "
                    f"the deletion of {ip.ip}: {e}"
                )
                return False
            else:
                session.commit()
                return True

    def get_ips_checking(self):
        with SessionLocal() as session:
            ips = []
            for ip in (
                session.query(IP).filter(IP.last_check_end < IP.last_check_start).all()
            ):
                ips.append(
                    IPModel(
                        id=ip.id,
                        ip=ip.ip,
                        last_check_start=ip.last_check_start,
                        last_check_end=ip.last_check_end,
                    )
                )
            return ips
